import * as http from 'http';
import * as https from 'https';
import { constants } from 'http2';
import { pluralize } from 'inflection';
import { BaseSchema, ObjectSchema } from 'valibot';
import { SerializerPair } from './serializers';
import {
	handleCreateItem, handleDeleteItem, handleEmplaceItem,
	handleGetCollection,
	handleGetItem,
	handleGetRoot,
	handleHasMethodAndUrl, handlePatchItem,
} from './handlers';
import Negotiator from 'negotiator';
import {getMethod, getUrl} from './utils';
import {EncodingPair} from './encodings';

export interface DataSource<T = object> {
	initialize(): Promise<unknown>;
	getMultiple(): Promise<T[]>;
	getSingle(id: string): Promise<T | null>;
	create(data: Partial<T>): Promise<T>;
	delete(id: string): Promise<unknown>;
	emplace(id: string, data: Partial<T>): Promise<[T, boolean]>;
	patch(id: string, data: Partial<T>): Promise<T | null>;
}

export interface ApplicationParams {
	name: string;
	dataSource?: (resource: Resource) => DataSource;
}

interface ResourceFactory {
	shouldCheckSerializersOnDelete(b: boolean): this;
	shouldThrow404OnDeletingNotFound(b: boolean): this;
	id(newIdAttr: string, params: IdParams): this;
	fullText(fullTextAttr: string): this;
	name(n: string): this;
	collection(n: string): this;
	route(n: string): this;
	allowFetchCollection(): this;
	allowFetchItem(): this;
	allowCreate(): this;
	allowPatch(): this;
	allowEmplace(): this;
	allowDelete(): this;
	revokeFetchCollection(): this;
	revokeFetchItem(): this;
	revokeCreate(): this;
	revokePatch(): this;
	revokeEmplace(): this;
	revokeDelete(): this;
}

export interface ResourceData<T extends BaseSchema> {
	idAttr: string;
	itemName?: string;
	collectionName?: string;
	routeName?: string;
	newId(dataSource: DataSource): string | number | unknown;
	schema: T;
	throws404OnDeletingNotFound: boolean;
	checksSerializersOnDelete: boolean;
	idSerializer: NonNullable<IdParams['serialize']>;
	idDeserializer: NonNullable<IdParams['deserialize']>;
}

export interface ResourcePermissions {
	canCreate: boolean;
	canFetchCollection: boolean;
	canFetchItem: boolean;
	canPatch: boolean;
	canEmplace: boolean;
	canDelete: boolean;
}

export type Resource<T extends BaseSchema = any> = ResourceData<T> & ResourceFactory & ResourcePermissions;

export interface ResourceWithDataSource<T extends BaseSchema = any> extends Resource<T> {
	dataSource: DataSource;
}

interface GenerationStrategy {
	(dataSource: DataSource, ...args: unknown[]): Promise<string | number | unknown>;
}

interface IdParams {
	generationStrategy: GenerationStrategy;
	serialize?: (id: unknown) => string;
	deserialize?: (id: string) => unknown;
}

const getAllowedMiddlewares = (resource: Resource, mainResourceId: string) => {
	const middlewares = [] as [string, Middleware][];
	if (mainResourceId === '') {
		if (resource.canFetchCollection) {
			middlewares.push(['GET', handleGetCollection]);
		}
		if (resource.canCreate) {
			middlewares.push(['POST', handleCreateItem]);
		}
		return middlewares;
	}

	if (resource.canFetchItem) {
		middlewares.push(['GET', handleGetItem]);
	}
	if (resource.canEmplace) {
		middlewares.push(['PUT', handleEmplaceItem]);
	}
	if (resource.canPatch) {
		middlewares.push(['PATCH', handlePatchItem]);
	}
	if (resource.canDelete) {
		middlewares.push(['DELETE', handleDeleteItem]);
	}

	return middlewares;
};

export const resource = <T extends BaseSchema>(schema: T): Resource<T> => {
	let theIdAttr: string;
	let theItemName: string;
	let theCollectionName: string;
	let theRouteName: string;
	let theIdGenerationStrategy: GenerationStrategy;
	let theIdSerializer: IdParams['serialize'];
	let theIdDeserializer: IdParams['deserialize'];
	let throw404OnDeletingNotFound = true;
	let checkSerializersOnDelete = false;
	const fullTextAttrs = new Set<string>();
	let canCreate = false;
	let canFetchCollection = false;
	let canFetchItem = false;
	let canPatch = false;
	let canEmplace = false;
	let canDelete = false;

	return {
		allowFetchCollection() {
			canFetchCollection = true;
			return this;
		},
		allowFetchItem() {
			canFetchItem = true;
			return this;
		},
		allowCreate() {
			canCreate = true;
			return this;
		},
		allowPatch() {
			canPatch = true;
			return this;
		},
		allowEmplace() {
			canEmplace = true;
			return this;
		},
		allowDelete() {
			canDelete = true;
			return this;
		},
		revokeFetchCollection() {
			canFetchCollection = false;
			return this;
		},
		revokeFetchItem() {
			canFetchItem = false;
			return this;
		},
		revokeCreate() {
			canCreate = false;
			return this;
		},
		revokePatch() {
			canPatch = false;
			return this;
		},
		revokeEmplace() {
			canEmplace = false;
			return this;
		},
		revokeDelete() {
			canDelete = false;
			return this;
		},
		get canCreate() {
			return canCreate;
		},
		get canFetchCollection() {
			return canFetchCollection;
		},
		get canFetchItem() {
			return canFetchItem;
		},
		get canPatch() {
			return canPatch;
		},
		get canEmplace() {
			return canEmplace;
		},
		get canDelete() {
			return canDelete;
		},
		shouldCheckSerializersOnDelete(b = true) {
			checkSerializersOnDelete = b;
			return this;
		},
		get checksSerializersOnDelete() {
			return checkSerializersOnDelete;
		},
		shouldThrow404OnDeletingNotFound(b = true) {
			throw404OnDeletingNotFound = b;
			return this;
		},
		get throws404OnDeletingNotFound() {
			return throw404OnDeletingNotFound;
		},
		get idSerializer() {
			return theIdSerializer;
		},
		get idDeserializer() {
			return theIdDeserializer;
		},
		id(newIdAttr: string, params: IdParams) {
			theIdAttr = newIdAttr;
			theIdGenerationStrategy = params.generationStrategy;
			theIdSerializer = params.serialize;
			theIdDeserializer = params.deserialize;
			return this;
		},
		newId(dataSource: DataSource) {
			return theIdGenerationStrategy(dataSource);
		},
		fullText(fullTextAttr: string) {
			if (
				schema.type === 'object'
				&& (
					schema as unknown as ObjectSchema<
						Record<string, BaseSchema>,
						undefined,
						Record<string, string>
					>
				)
					.entries[fullTextAttr]?.type === 'string'
			) {
				fullTextAttrs.add(fullTextAttr);
				return this;
			}

			throw new Error(`Could not set attribute ${fullTextAttr} as fulltext.`);
		},
		name(n: string) {
			theItemName = n;
			theCollectionName = theCollectionName ?? pluralize(theItemName).toLowerCase();
			theRouteName = theRouteName ?? theCollectionName;
			return this;
		},
		collection(n: string) {
			theCollectionName = n;
			theRouteName = theRouteName ?? theCollectionName;
			return this;
		},
		route(n: string) {
			theRouteName = n;
			return this;
		},
		get idAttr() {
			return theIdAttr;
		},
		get collectionName() {
			return theCollectionName;
		},
		get itemName() {
			return theItemName;
		},
		get routeName() {
			return theRouteName;
		},
		get schema() {
			return schema;
		}
	} as Resource;
};

interface CreateServerParams {
	baseUrl?: string;
	host?: string;
	cert?: string;
	key?: string;
	requestTimeout?: number;
}

type RequestListenerWithReturn<
	P extends unknown = unknown,
	Q extends typeof http.IncomingMessage = typeof http.IncomingMessage,
	R extends typeof http.ServerResponse = typeof http.ServerResponse
> = (
	...args: Parameters<http.RequestListener<Q, R>>
) => P;

interface HandlerState {
	handled: boolean;
}

export interface ApplicationState {
	resources: Set<ResourceWithDataSource>;
	serializers: Map<string, SerializerPair>;
	encodings: Map<string, EncodingPair>;
}

interface MiddlewareArgs {
	handlerState: HandlerState;
	appState: ApplicationState;
	appParams: ApplicationParams;
	serverParams: CreateServerParams;
	requestBodyEncodingPair: EncodingPair;
	responseBodySerializerPair: SerializerPair;
	responseMediaType: string;
	resource: ResourceWithDataSource;
	resourceId: string;
	query: URLSearchParams;
}

export interface Middleware {
	(args: MiddlewareArgs): RequestListenerWithReturn<HandlerState | Promise<HandlerState>>
}

export interface Application {
	contentType(mimeTypePrefix: string, serializerPair: SerializerPair): this;
	encoding(encoding: string, encodingPair: EncodingPair): this;
	resource(resRaw: Partial<Resource>): this;
	createServer(serverParams?: CreateServerParams): http.Server | https.Server;
}

export const application = (appParams: ApplicationParams): Application => {
	const appState: ApplicationState = {
		resources: new Set<ResourceWithDataSource>(),
		serializers: new Map<string, SerializerPair>(),
		encodings: new Map<string, EncodingPair>()
	};

	return {
		contentType(mimeTypePrefix: string, serializerPair: SerializerPair) {
			appState.serializers.set(mimeTypePrefix, serializerPair);
			return this;
		},
		encoding(encoding: string, encodingPair: EncodingPair) {
			appState.encodings.set(encoding, encodingPair);
			return this;
		},
		resource(resRaw: Partial<Resource>) {
			const res = resRaw as Partial<ResourceWithDataSource>;
			res.dataSource = res.dataSource ?? appParams.dataSource?.(res as Resource);
			if (typeof res.dataSource === 'undefined') {
				throw new Error(`Resource ${res.itemName} must have a data source.`);
			}
			appState.resources.add(res as ResourceWithDataSource);
			return this;
		},
		createServer(serverParams = {} as CreateServerParams) {
			const server = 'key' in serverParams && 'cert' in serverParams
				? https.createServer({
					key: serverParams.key,
					cert: serverParams.cert,
					requestTimeout: serverParams.requestTimeout
				})
				: http.createServer({
					requestTimeout: serverParams.requestTimeout
				});

			server.on('request', async (req, res) => {
				const method = getMethod(req);
				const baseUrl = serverParams.baseUrl ?? '';
				const { url, query } = getUrl(req, baseUrl);

				const negotiator = new Negotiator(req);
				const availableMediaTypes = Array.from(appState.serializers.keys());
				const responseMediaType = negotiator.mediaType(availableMediaTypes);

				if (typeof responseMediaType === 'undefined') {
					res.statusCode = constants.HTTP_STATUS_NOT_ACCEPTABLE;
					res.end();
					return;
				}

				const responseBodySerializerPair = appState.serializers.get(responseMediaType);
				if (typeof responseBodySerializerPair === 'undefined') {
					res.statusCode = constants.HTTP_STATUS_NOT_ACCEPTABLE;
					res.end();
					return;
				}

				const availableEncodings = Array.from(appState.encodings.keys());
				const encoding = negotiator.encoding(availableEncodings);

				if (typeof encoding === 'undefined') {
					res.statusCode = constants.HTTP_STATUS_NOT_ACCEPTABLE;
					res.end();
					return;
				}

				const requestBodyEncodingPair = appState.encodings.get(encoding);
				if (typeof requestBodyEncodingPair === 'undefined') {
					res.statusCode = constants.HTTP_STATUS_NOT_ACCEPTABLE;
					res.end();
					return;
				}

				const middlewareArgs: Omit<MiddlewareArgs, 'resource' | 'resourceId'> = {
					handlerState: {
						handled: false
					},
					appState,
					appParams,
					serverParams,
					responseBodySerializerPair,
					responseMediaType,
					query,
					requestBodyEncodingPair,
				};

				const methodAndUrl = await handleHasMethodAndUrl(middlewareArgs as MiddlewareArgs)(req, res);
				if (methodAndUrl.handled) {
					return;
				}

				if (url === '/') {
					const middlewareState = await handleGetRoot(middlewareArgs as MiddlewareArgs)(req, res);
					if (middlewareState.handled) {
						return;
					}

					res.writeHead(constants.HTTP_STATUS_METHOD_NOT_ALLOWED, {
						Allow: 'HEAD, GET'
					});
					res.end();
					return;
				}

				const [, resourceRouteName, resourceId = ''] = url.split('/');
				const resource = Array.from(appState.resources).find((r) => r.routeName === resourceRouteName);
				if (typeof resource === 'undefined') {
					res.statusCode = constants.HTTP_STATUS_NOT_FOUND;
					res.statusMessage = 'URL Not Found';
					res.end();
					return;
				}

				const middlewares = getAllowedMiddlewares(resource, resourceId);
				const middlewareState = await middlewares
					.reduce(
						async (currentHandlerStatePromise, [middlewareMethod, middleware]) => {
							const currentHandlerState = await currentHandlerStatePromise;
							if (method !== middlewareMethod) {
								return currentHandlerState;
							}

							if (currentHandlerState.handled) {
								return currentHandlerState;
							}

							return middleware({
								...middlewareArgs,
								handlerState: currentHandlerState,
								resource,
								resourceId: resourceId,
							})(req, res);
						},
						Promise.resolve<HandlerState>({
							handled: false
						})
					);

				if (middlewareState.handled) {
					return;
				}

				if (middlewares.length > 0) {
					res.writeHead(constants.HTTP_STATUS_METHOD_NOT_ALLOWED, {
						Allow: middlewares.map((m) => m[0]).join(', ')
					});
					res.end();
					return;
				}

				res.statusCode = constants.HTTP_STATUS_NOT_FOUND;
				res.statusMessage = 'URL Not Found';
				res.end();
				return;
			});

			return server;
		}
	};
};
