import { constants } from 'http2';
import * as v from 'valibot';
import {Middleware} from './core';
import {getBody, getDeserializerObjects, getMethod, getUrl} from './utils';
import {IncomingMessage, ServerResponse} from 'http';

export const handleHasMethodAndUrl: Middleware = ({}) => (req: IncomingMessage, res: ServerResponse) => {
	if (!req.method) {
		res.writeHead(constants.HTTP_STATUS_METHOD_NOT_ALLOWED, {
			'Allow': 'HEAD, GET, POST, PUT, PATCH, DELETE'
		});
		res.end();
		return {
			handled: true
		};
	}

	if (!req.url) {
		res.statusCode = constants.HTTP_STATUS_BAD_REQUEST;
		res.end();
		return {
			handled: true
		};
	}

	return {
		handled: false
	};
};

export const handleGetRoot: Middleware = ({
	appState,
	appParams,
	responseBodySerializerPair,
	serverParams,
	responseMediaType,
}) => (_req: IncomingMessage, res: ServerResponse) => {
	const singleResDatum = {
		name: appParams.name
	};
	const theFormatted = responseBodySerializerPair.serialize(singleResDatum);
	res.writeHead(constants.HTTP_STATUS_OK, {
		'Content-Type': responseMediaType,
		// we are using custom headers for links because the standard Link header
		// is referring to the document metadata (e.g. author, next page, etc)
		// https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Link
		'X-Resource-Link': Array.from(appState.resources)
			.map((r) =>
				`<${serverParams.baseUrl}/${r.routeName}>; name="${r.collectionName}"`,
			)
			.join(', ')
	});
	res.end(theFormatted);
	return {
		handled: true
	};
};

export const handleGetCollection: Middleware = ({
	appState,
	serverParams,
	responseBodySerializerPair,
	responseMediaType,
}) => async (req: IncomingMessage, res: ServerResponse) => {
	const baseUrl = serverParams.baseUrl ?? '';
	const { url } = getUrl(req, baseUrl);

	const [, mainResourceRouteName, mainResourceId = ''] = url.split('/');
	if (mainResourceId !== '') {
		return {
			handled: false
		}
	}

	const theResource = Array.from(appState.resources).find((r) => r.routeName === mainResourceRouteName);
	if (typeof theResource === 'undefined') {
		return {
			handled: false
		};
	}

	try {
		await theResource.dataSource.initialize();
		// TODO querying mechanism
		const resData = await theResource.dataSource.getMultiple(); // TODO paginated responses per resource
		const theFormatted = responseBodySerializerPair.serialize(resData);

		res.writeHead(constants.HTTP_STATUS_OK, {
			'Content-Type': responseMediaType,
			'X-Resource-Total-Item-Count': resData.length
		});
		res.end(theFormatted);
	} catch {
		res.statusCode = constants.HTTP_STATUS_INTERNAL_SERVER_ERROR;
		res.end();
	}

	return {
		handled: true
	};
};

export const handleGetItem: Middleware = ({
	appState,
	serverParams,
	responseBodySerializerPair,
	responseMediaType,
}) => async (req: IncomingMessage, res: ServerResponse) => {
	const method = getMethod(req);
	if (method !== 'GET') {
		return {
			handled: false
		};
	}

	const baseUrl = serverParams.baseUrl ?? '';
	const { url } = getUrl(req, baseUrl);

	const [, mainResourceRouteName, mainResourceId = ''] = url.split('/');
	if (mainResourceId === '') {
		return {
			handled: false
		}
	}

	const theResource = Array.from(appState.resources).find((r) => r.routeName === mainResourceRouteName);
	if (typeof theResource === 'undefined') {
		return {
			handled: false
		};
	}

	try {
		await theResource.dataSource.initialize();
		const singleResDatum = await theResource.dataSource.getSingle(mainResourceId);
		if (singleResDatum) {
			const theFormatted = responseBodySerializerPair.serialize(singleResDatum);
			res.writeHead(constants.HTTP_STATUS_OK, {'Content-Type': responseMediaType});
			res.end(theFormatted);
			return {
				handled: true
			};
		}

		res.statusCode = constants.HTTP_STATUS_NOT_FOUND;
		res.statusMessage = `${theResource.itemName} Not Found`;
		res.end();
		return {
			handled: true
		};
	} catch {
		res.statusCode = constants.HTTP_STATUS_INTERNAL_SERVER_ERROR;
		res.end();
		return {
			handled: true
		};
	}
};


export const handleDeleteItem: Middleware = ({
	resource,
	resourceId,
}) => async (_req: IncomingMessage, res: ServerResponse) => {
	try {
		await resource.dataSource.initialize();
		const response = await resource.dataSource.delete(resourceId);
		if (typeof response !== 'undefined' && !response && resource.throws404OnDeletingNotFound) {
			res.statusCode = constants.HTTP_STATUS_NOT_FOUND;
			res.statusMessage = `${resource.itemName} Not Found`;
		} else {
			res.statusCode = constants.HTTP_STATUS_NO_CONTENT;
		}
		res.end();
		return {
			handled: true
		};
	} catch {
		// TODO error handling
		// what if item is already deleted? Should we hide it by returning no content or throw a 404?
	}

	res.statusCode = constants.HTTP_STATUS_INTERNAL_SERVER_ERROR;
	res.end();
	return {
		handled: true
	};
};

export const handlePatchItem: Middleware = ({
	appState,
	responseBodySerializerPair,
	responseMediaType,
	resource,
	resourceId,
}) => async (req: IncomingMessage, res: ServerResponse) => {
	const { deserializerPair: requestBodyDeserializerPair, encodingPair: requestBodyEncodingPair } = getDeserializerObjects(appState, req);
	if (typeof requestBodyDeserializerPair === 'undefined' || typeof requestBodyEncodingPair === 'undefined') {
		res.statusCode = constants.HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE;
		res.end();
		return {
			handled: true
		};
	}

	await resource.dataSource.initialize();
	const existing = await resource.dataSource.getSingle(resourceId);
	if (!existing) {
		res.statusCode = constants.HTTP_STATUS_NOT_FOUND;
		res.statusMessage = `${resource.itemName} Not Found`;
		res.end();
		return {
			handled: true
		};
	}

	let bodyDeserialized: unknown;
	try {
		const schema = resource.schema.type === 'object' ? resource.schema as v.ObjectSchema<any> : resource.schema
		bodyDeserialized = await getBody(
			req,
			requestBodyDeserializerPair,
			requestBodyEncodingPair,
			schema.type === 'object'
				? v.partial(
					schema as v.ObjectSchema<any>,
					(schema as v.ObjectSchema<any>).rest,
					(schema as v.ObjectSchema<any>).pipe
				)
				: schema
		);
	} catch (errRaw) {
		const err = errRaw as v.ValiError;
		res.statusCode = constants.HTTP_STATUS_BAD_REQUEST;
		res.statusMessage = `Invalid ${resource.itemName}`;

		if (Array.isArray(err.issues)) {
			// TODO better error reporting, localizable messages
			const theFormatted = responseBodySerializerPair.serialize(
				err.issues.map((i) => (
					`${i.path?.map((p) => p.key)?.join('.') ?? i.reason}: ${i.message}`
				))
			);
			res.end(theFormatted);
		} else {
			res.end();
		}
		return {
			handled: true
		};
	}

	try {
		const params = bodyDeserialized as Record<string, unknown>;
		await resource.dataSource.initialize();
		const newObject = await resource.dataSource.patch(resourceId, params);
		const theFormatted = responseBodySerializerPair.serialize(newObject);
		res.writeHead(constants.HTTP_STATUS_OK, {
			'Content-Type': responseMediaType,
		});
		res.end(theFormatted);
	} catch {
		res.statusCode = constants.HTTP_STATUS_INTERNAL_SERVER_ERROR;
		res.statusMessage = `Could Not Return ${resource.itemName}`;
		res.end();
	}
	return {
		handled: true
	};
};

export const handleCreateItem: Middleware = ({
	appState,
	serverParams,
	responseMediaType,
	responseBodySerializerPair,
	resource,
}) => async (req: IncomingMessage, res: ServerResponse) => {
	const { deserializerPair: requestBodyDeserializerPair, encodingPair: requestBodyEncodingPair } = getDeserializerObjects(appState, req);
	if (typeof requestBodyDeserializerPair === 'undefined' || typeof requestBodyEncodingPair === 'undefined') {
		res.statusCode = constants.HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE;
		res.end();
		return {
			handled: true
		};
	}

	let bodyDeserialized: unknown;
	try {
		bodyDeserialized = await getBody(req, requestBodyDeserializerPair, requestBodyEncodingPair, resource.schema);
	} catch (errRaw) {
		const err = errRaw as v.ValiError;
		res.statusCode = constants.HTTP_STATUS_BAD_REQUEST;
		res.statusMessage = `Invalid ${resource.itemName}`;

		if (Array.isArray(err.issues)) {
			// TODO better error reporting, localizable messages
			const theFormatted = responseBodySerializerPair.serialize(
				err.issues.map((i) => (
					`${i.path?.map((p) => p.key)?.join('.') ?? i.reason}: ${i.message}`
				))
			);
			res.end(theFormatted);
		} else {
			res.end();
		}
		return {
			handled: true
		};
	}

	try {
		await resource.dataSource.initialize();
		const newId = await resource.newId(resource.dataSource);
		const params = bodyDeserialized as Record<string, unknown>;
		params[resource.idAttr] = newId;
		const newObject = await resource.dataSource.create(params);
		const theFormatted = responseBodySerializerPair.serialize(newObject);
		res.writeHead(constants.HTTP_STATUS_OK, {
			'Content-Type': responseMediaType,
			'Location': `${serverParams.baseUrl}/${resource.routeName}/${newId}`
		});
		res.end(theFormatted);
	} catch {
		res.statusCode = constants.HTTP_STATUS_INTERNAL_SERVER_ERROR;
		res.statusMessage = `Could Not Return ${resource.itemName}`;
		res.end();
	}
	return {
		handled: true
	};
}

export const handleEmplaceItem: Middleware = ({
	appState,
	serverParams,
	responseBodySerializerPair,
	responseMediaType,
	resource,
	resourceId,
}) => async (req: IncomingMessage, res: ServerResponse) => {
	const { deserializerPair: requestBodyDeserializerPair, encodingPair: requestBodyEncodingPair } = getDeserializerObjects(appState, req);
	if (typeof requestBodyDeserializerPair === 'undefined' || typeof requestBodyEncodingPair === 'undefined') {
		res.statusCode = constants.HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE;
		res.end();
		return {
			handled: true
		};
	}

	let bodyDeserialized: unknown;
	try {
		const schema = resource.schema.type === 'object' ? resource.schema as v.ObjectSchema<any> : resource.schema
		//console.log(schema);
		bodyDeserialized = await getBody(
			req,
			requestBodyDeserializerPair,
			requestBodyEncodingPair,
			schema.type === 'object'
				? v.merge([
					schema as v.ObjectSchema<any>,
					v.object({
						[resource.idAttr]: v.transform(
							v.any(),
							input => resource.idSerializer(input),
							v.literal(resourceId)
						)
					})
				])
				: schema
		);
	} catch (errRaw) {
		const err = errRaw as v.ValiError;
		res.statusCode = constants.HTTP_STATUS_BAD_REQUEST;
		res.statusMessage = `Invalid ${resource.itemName}`;

		if (Array.isArray(err.issues)) {
			// TODO better error reporting, localizable messages
			const theFormatted = responseBodySerializerPair.serialize(
				err.issues.map((i) => (
					`${i.path?.map((p) => p.key)?.join('.') ?? i.reason}: ${i.message}`
				))
			);
			res.end(theFormatted);
		} else {
			res.end();
		}
		return {
			handled: true
		};
	}

	try {
		await resource.dataSource.initialize();
		const params = bodyDeserialized as Record<string, unknown>;
		const [newObject, isCreated] = await resource.dataSource.emplace(resourceId, params);
		const theFormatted = responseBodySerializerPair.serialize(newObject);
		if (isCreated) {
			res.writeHead(constants.HTTP_STATUS_CREATED, {
				'Content-Type': responseMediaType,
				'Location': `${serverParams.baseUrl}/${resource.routeName}/${resourceId}`
			});
		} else {
			res.writeHead(constants.HTTP_STATUS_OK, {
				'Content-Type': responseMediaType,
			});
		}
		res.end(theFormatted);
	} catch {
		res.statusCode = constants.HTTP_STATUS_INTERNAL_SERVER_ERROR;
		res.statusMessage = `Could Not Return ${resource.itemName}`;
		res.end();
	}
	return {
		handled: true
	};
}
